from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MismatchedTypes
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest


class TestDict(BoaTest):

    def test_dict_int_keys(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.NEWMAP  # a = {1: 15, 2: 14, 3: 13}
            + Opcode.DUP
            + Opcode.PUSH1      # map[1] = 15
            + Opcode.PUSH15
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH2      # map[2] = 14
            + Opcode.PUSH14
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH3      # map[3] = 13
            + Opcode.PUSH13
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/dict_test/IntKeyDict.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_str_keys(self):
        one = String('one').to_bytes()
        two = String('two').to_bytes()
        three = String('three').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.NEWMAP  # a = {'one': 1, 'two': 2, 'three': 3}
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['one'] = 1
            + Integer(len(one)).to_byte_array(min_length=1)
            + one
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['two'] = 2
            + Integer(len(two)).to_byte_array(min_length=1)
            + two
            + Opcode.PUSH2
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['three'] = 3
            + Integer(len(three)).to_byte_array(min_length=1)
            + three
            + Opcode.PUSH3
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/dict_test/StrKeyDict.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_any_value(self):
        nine = String('nine').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.NEWMAP  # a = {1: True, 2: 4, 3: 'nine'}
            + Opcode.DUP
            + Opcode.PUSH1      # map[1] = True
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH2      # map[2] = 4
            + Opcode.PUSH4
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH3      # map[3] = 'nine'
            + Opcode.PUSHDATA1
            + Integer(len(nine)).to_byte_array(min_length=1)
            + nine
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/dict_test/AnyValueDict.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_of_dicts(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.NEWMAP  # a = map
            + Opcode.DUP
            + Opcode.PUSH1      # map[1] = {14: False, 12: True, 5: True}
                + Opcode.NEWMAP
                + Opcode.DUP
                + Opcode.PUSH14
                + Opcode.PUSH0
                + Opcode.SETITEM
                + Opcode.DUP
                + Opcode.PUSH12
                + Opcode.PUSH1
                + Opcode.SETITEM
                + Opcode.DUP
                + Opcode.PUSH5
                + Opcode.PUSH1
                + Opcode.SETITEM
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH2      # map[2] = {0: True, 6: False}
                + Opcode.NEWMAP
                + Opcode.DUP
                + Opcode.PUSH0
                + Opcode.PUSH1
                + Opcode.SETITEM
                + Opcode.DUP
                + Opcode.PUSH6
                + Opcode.PUSH0
                + Opcode.SETITEM
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH3      # map[3] = {11: False}
                + Opcode.NEWMAP
                + Opcode.DUP
                + Opcode.PUSH11
                + Opcode.PUSH0
                + Opcode.SETITEM
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/dict_test/DictOfDict.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_assign_empty_dict(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.NEWMAP  # a = {}
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/dict_test/EmptyDictAssignment.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_type_hint_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.NEWMAP  # a = {1: 15, 2: 14, 3: 13}
            + Opcode.DUP
            + Opcode.PUSH1      # map[1] = 15
            + Opcode.PUSH15
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH2      # map[2] = 14
            + Opcode.PUSH14
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH3      # map[3] = 13
            + Opcode.PUSH13
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/dict_test/TypeHintAssignment.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_variable_keys_and_values(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x04'
            + b'\x00'
            + Opcode.PUSH1   # a = 1
            + Opcode.STLOC0
            + Opcode.PUSH2   # b = 2
            + Opcode.STLOC1
            + Opcode.PUSH3   # c = 3
            + Opcode.STLOC2
            + Opcode.NEWMAP  # d = {a: c, b: a, c: b}
            + Opcode.DUP
            + Opcode.LDLOC0     # map[a] = c
            + Opcode.LDLOC2
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.LDLOC1     # map[b] = a
            + Opcode.LDLOC0
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.LDLOC2     # map[c] = b
            + Opcode.LDLOC1
            + Opcode.SETITEM
            + Opcode.STLOC3
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/dict_test/VariableDict.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_get_value(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.PUSH0
            + Opcode.PICKITEM
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/dict_test/GetValue.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_get_value_mismatched_type(self):
        path = '%s/boa3_test/test_sc/dict_test/MismatchedTypeGetValue.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_dict_set_value(self):
        ok = String('ok').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.PUSH0
            + Opcode.PUSHDATA1
            + Integer(len(ok)).to_byte_array(min_length=1)
            + ok
            + Opcode.SETITEM
            + Opcode.LDARG0
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/dict_test/SetValue.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_set_value_mismatched_type(self):
        path = '%s/boa3_test/test_sc/dict_test/MismatchedTypeSetValue.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_dict_keys(self):
        one = String('one').to_bytes()
        two = String('two').to_bytes()
        three = String('three').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.NEWMAP  # a = {'one': 1, 'two': 2, 'three': 3}
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['one'] = 1
            + Integer(len(one)).to_byte_array(min_length=1)
            + one
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['two'] = 2
            + Integer(len(two)).to_byte_array(min_length=1)
            + two
            + Opcode.PUSH2
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['three'] = 3
            + Integer(len(three)).to_byte_array(min_length=1)
            + three
            + Opcode.PUSH3
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.LDLOC0  # b = a.keys()
            + Opcode.KEYS
            + Opcode.STLOC1
            + Opcode.LDLOC1  # return b
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/dict_test/KeysDict.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_keys_mismatched_type(self):
        path = '%s/boa3_test/test_sc/dict_test/MismatchedTypeKeysDict.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_dict_values(self):
        one = String('one').to_bytes()
        two = String('two').to_bytes()
        three = String('three').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.NEWMAP  # a = {'one': 1, 'two': 2, 'three': 3}
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['one'] = 1
            + Integer(len(one)).to_byte_array(min_length=1)
            + one
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['two'] = 2
            + Integer(len(two)).to_byte_array(min_length=1)
            + two
            + Opcode.PUSH2
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['three'] = 3
            + Integer(len(three)).to_byte_array(min_length=1)
            + three
            + Opcode.PUSH3
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.LDLOC0  # b = a.values()
            + Opcode.VALUES
            + Opcode.STLOC1
            + Opcode.LDLOC1  # return b
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/dict_test/ValuesDict.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_values_mismatched_type(self):
        path = '%s/boa3_test/test_sc/dict_test/MismatchedTypeValuesDict.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)



from __future__ import annotations

import abc
from contextlib import suppress
from copy import deepcopy
from enum import Enum, auto
from typing import Optional, Iterator, Tuple, TypeVar, Any

from boa3.neo3 import storage
from boa3.neo3.core import types
from boa3.neo3.network import payloads

TKey = TypeVar('TKey', bound='serialization.ISerializable')
TValue = TypeVar('TValue', bound='serialization.ISerializable')


class TrackState(Enum):
    NONE = auto()
    ADDED = auto()
    CHANGED = auto()
    DELETED = auto()


class Trackable:
    def __init__(self, key: TKey, value: TValue, state: TrackState):
        self.key = key
        self.item = value
        self.state = state


class CachedAccess:
    def __init__(self, db):
        self._dictionary = {}
        self._db = db
        self._internal_get = None
        self._internal_try_get = None
        self._internal_all = None

    def __getitem__(self, key):
        trackable = self._dictionary.get(key, None)  # type: Trackable
        if trackable is not None:
            if trackable.state == TrackState.DELETED:
                raise KeyError
        else:
            trackable = Trackable(key, self._internal_get(key), TrackState.NONE)
            self._dictionary.update({key: trackable})
        return trackable.item

    def _put(self, key, value):
        trackable = self._dictionary.get(key, None)  # type: Trackable
        if trackable is not None:
            if trackable.state != TrackState.DELETED:
                raise ValueError("Value already exists")
        else:
            if self._internal_try_get(key) is not None:
                raise ValueError("Value already exists")

        if trackable is None:
            trackable = Trackable(key, value, TrackState.ADDED)
        else:
            trackable = Trackable(key, value, TrackState.CHANGED)

        self._dictionary.update({key: trackable})

    def _get(self, key, read_only=False):
        """
        Return the value stored with `key`.

        Args:
            key: the identifier that can be used to retrieve the stored value
            read_only:

        Raises:
            KeyError if the key is not found.
        """
        value = self[key]
        if value is None:
            raise KeyError

        if read_only:
            # make sure that even trackables in ADDED/CHANGED state can't be modified
            return deepcopy(value)

        if self._dictionary[key].state == TrackState.NONE:
            self._dictionary[key].state = TrackState.CHANGED
        return value

    def _delete(self, key) -> None:
        """
        Delete the value stored under `key`.

        Args:
            key: the identifier that can be used to identify the stored value
        """
        trackable = self._dictionary.get(key, None)  # type: Trackable
        if trackable is not None:
            if trackable.state == TrackState.ADDED:
                self._dictionary.pop(key)
            else:
                trackable.state = TrackState.DELETED
        else:
            item = self._internal_try_get(key)
            if item is None:
                return

            self._dictionary.update({key: Trackable(key, item, TrackState.DELETED)})

    @abc.abstractmethod
    def commit(self):
        """ Persist changes"""

    @abc.abstractmethod
    def create_snapshot(self):
        """ Deep copy. """


class AttributeCache:
    def __init__(self):
        self._item = None
        self._state = TrackState.NONE

    def put(self, item) -> None:
        self._item = item

    def get(self, read_only=False) -> Any:
        if self._item is None:
            self._item = self._get_internal()

        if read_only:
            return deepcopy(self._item)
        else:
            self._state = TrackState.CHANGED
            return self._item

    def commit(self) -> None:
        if self._state == TrackState.CHANGED:
            self._update_internal(self._item)

    def create_snapshot(self) -> CloneAttributeCache:
        return CloneAttributeCache(self)

    @abc.abstractmethod
    def _get_internal(self):
        """ Return the value from the real backend. """

    @abc.abstractmethod
    def _update_internal(self, value):
        """ Update the value in the real backend. """


class CachedBlockAccess(CachedAccess):
    def __init__(self, db):
        super(CachedBlockAccess, self).__init__(db)
        self._height_hash_mapping = {}
        self._internal_get = self._db._internal_block_get
        self._internal_try_get = self._db._internal_block_try_get
        self._internal_all = self._db._internal_block_all

    def put(self, block: payloads.Block) -> None:
        """
        Store a block.

        Args:
            block: instance.

        Raises:
            ValueError: if a duplicate item is found.
        """
        block_hash = block.hash()
        super(CachedBlockAccess, self)._put(block_hash, block)
        self._height_hash_mapping.update({block.index: block_hash})

    def get(self, hash: types.UInt256, read_only=False) -> payloads.Block:
        """
        Retrieve a block.

        Args:
            hash: block hash.
            read_only: set to True to safeguard against return value modifications being persisted when committing.

        Raises:
            KeyError: if the item is not found.
        """
        block = super(CachedBlockAccess, self)._get(hash, read_only)
        self._height_hash_mapping.update({block.index: block.hash()})
        return block

    def get_by_height(self, height: int, read_only=False) -> payloads.Block:
        """
        Retrieve a block by its height.

        Args:
            height: target block index/height.
            read_only: set to True to safeguard against return value modifications being persisted when committing.

        Raises:
            KeyError: if the item is not found.
        """
        block_hash = self._height_hash_mapping.get(height, None)
        if block_hash is None:
            raise KeyError
        return self.get(block_hash, read_only)

    def try_get(self, hash: types.UInt256, read_only=False) -> Optional[payloads.Block]:
        """
        Try to retrieve a block.

        Args:
            hash: block hash.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        """
        try:
            return self.get(hash, read_only)
        except KeyError:
            return None

    def try_get_by_height(self, height, read_only=False) -> Optional[payloads.Block]:
        """
        Try to retrieve a block.

        Args:
            hash: block hash.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        """
        block_hash = self._height_hash_mapping.get(height, None)
        if block_hash is None:
            return None

        return self.try_get(block_hash, read_only)

    def delete(self, hash: types.UInt256) -> None:
        """
        Remove a block.

        Args:
            hash: block hash.
        """
        super(CachedBlockAccess, self)._delete(hash)
        with suppress(KeyError):
            self._height_hash_mapping.pop(hash)

    def all(self) -> Iterator[payloads.Block]:
        """
        Retrieve all blocks (readonly)
        """
        blocks = []
        for block in self._internal_all():
            if block.hash() not in self._dictionary:
                blocks.append(block)

        for k, v in self._dictionary.items():
            if v.state != TrackState.DELETED:
                blocks.append(deepcopy(v.item))

        blocks.sort(key=lambda block: block.hash().to_array())
        for block in blocks:
            yield block
        return None


class CachedContractAccess(CachedAccess):
    def __init__(self, db):
        super(CachedContractAccess, self).__init__(db)
        self._internal_get = self._db._internal_contract_get
        self._internal_try_get = self._db._internal_contract_try_get
        self._internal_all = self._db._internal_contract_all

    def put(self, contract: storage.ContractState) -> None:
        """
        Store a contract.

        Args:
            contract: contract state instance.

        Raises:
            ValueError: if a duplicate item is found.
        """
        super(CachedContractAccess, self)._put(contract.script_hash(), contract)

    def get(self, script_hash: types.UInt160, read_only=False) -> storage.ContractState:
        """
        Retrieve a contract.

        Args:
            script_hash: contract script hash.
            read_only: set to True to safeguard against return value modifications being persisted when committing.

        Raises:
            KeyError: if the item is not found.
        """
        return super(CachedContractAccess, self)._get(script_hash, read_only)

    def try_get(self, script_hash: types.UInt160, read_only=False) -> Optional[storage.ContractState]:
        """
        Try to retrieve a contract.

        Args:
            script_hash: contract script hash.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        """
        try:
            return self.get(script_hash, read_only)
        except KeyError:
            return None

    def delete(self, script_hash: types.UInt160) -> None:
        """
        Remove a transaction.

        Args:
            script_hash: contract script hash.
        """
        super(CachedContractAccess, self)._delete(script_hash)

    def all(self) -> Iterator[storage.ContractState]:
        """
        Retrieve all contracts (readonly)
        """
        contracts = []
        for contract in self._internal_all():
            if contract.script_hash() not in self._dictionary:
                contracts.append(contract)

        for k, v in self._dictionary.items():
            if v.state != TrackState.DELETED:
                contracts.append(deepcopy(v.item))

        contracts.sort(key=lambda contract: contract.script_hash().to_array())
        for contract in contracts:
            yield contract
        return None


class CachedStorageAccess(CachedAccess):
    def __init__(self, db):
        super(CachedStorageAccess, self).__init__(db)
        self._internal_get = self._db._internal_storage_get
        self._internal_try_get = self._db._internal_storage_try_get
        self._internal_all = self._db._internal_storage_all
        self._internal_find = self._db._internal_storage_find

    def put(self, key: storage.StorageKey, value: storage.StorageItem) -> None:
        """
        Store the value under the given key.

        Args:
            key: identifier to store the value under.
            value: the value to be persisted.
        """
        super(CachedStorageAccess, self)._put(key, value)

    def get(self, key: storage.StorageKey, read_only=False) -> storage.StorageItem:
        """
        Retrieve a value from storage.

        Args:
            key: identifier.
            read_only: set to True to safeguard against return value modifications being persisted when committing.

        Raises:
            KeyError: if the item is not found.
        """
        return super(CachedStorageAccess, self)._get(key, read_only)

    def try_get(self, key: storage.StorageKey, read_only=False) -> Optional[storage.StorageItem]:
        """
        Try to retrieve a value from storage.

        Args:
            key: identifier.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        """
        try:
            return self.get(key, read_only)
        except KeyError:
            return None

    def delete(self, key: storage.StorageKey) -> None:
        """
        Remove a key/value pair from storage

        Args:
            key: identifier to locate value.
        """
        super(CachedStorageAccess, self)._delete(key)

    def all(self, contract_script_hash: types.UInt160 = None) -> Iterator[Tuple[storage.StorageKey,
                                                                                storage.StorageItem]]:
        """
        Retrieve all storage key/value pairs, sorted by key (readonly).

        Note:
            Return values are sorted to give deterministic behaviour in smart contracts.

        Args:
            contract_script_hash: smart contract script hash to limit results to. If not specified, returns for all
            contracts.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        """
        pairs = []
        for k, v in self._internal_all(contract_script_hash):
            if k not in self._dictionary:
                pairs.append((k, v))

        for k, v in self._dictionary.items():
            if v.state != TrackState.DELETED:
                pairs.append((deepcopy(v.key), deepcopy(v.item)))

        pairs.sort(key=lambda keypair: keypair[0].to_array())
        for pair in pairs:
            yield pair
        return None

    def find(self, contract_script_hash: types.UInt160, key_prefix: bytes) -> Iterator[Tuple[storage.StorageKey,
                                                                                             storage.StorageItem]]:
        """
        Retrieve all storage key/value pairs (readonly).

        Args:
            contract_script_hash: script hash of smart contract to search storage of.
            key_prefix: the prefix part of the storage.StorageKey.key to look for.

        """
        pairs = []
        for k, v in self._internal_find(contract_script_hash, key_prefix):
            if k not in self._dictionary:
                pairs.append((k, v))

        for k, v in self._dictionary.items():
            if v.state != TrackState.DELETED and k.key.startswith(key_prefix):
                pairs.append((deepcopy(v.key), deepcopy(v.item)))

        pairs.sort(key=lambda keypair: keypair[0].to_array())
        for pair in pairs:
            yield pair
        return None


class CachedTXAccess(CachedAccess):
    def __init__(self, db):
        super(CachedTXAccess, self).__init__(db)
        self._internal_get = self._db._internal_transaction_get
        self._internal_try_get = self._db._internal_transaction_try_get
        self._internal_all = self._db._internal_transaction_all

    def put(self, tx: payloads.Transaction) -> None:
        """
        Store a transaction.

        Args:
            tx: instance.

        Raises:
            ValueError: if a duplicate item is found.
        """
        super(CachedTXAccess, self)._put(tx.hash(), tx)

    def get(self, hash: types.UInt256, read_only=False) -> payloads.Transaction:
        """
        Retrieve a transaction.

        Args:
            hash: transaction hash.
            read_only: set to True to safeguard against return value modifications being persisted when committing.

        Raises:
            KeyError: if the item is not found.
        """
        return super(CachedTXAccess, self)._get(hash, read_only)

    def try_get(self, hash: types.UInt256, read_only=False) -> Optional[payloads.Transaction]:
        """
        Try to retrieve a transaction.

        Args:
            hash: transaction hash.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        """
        try:
            return self.get(hash, read_only)
        except KeyError:
            return None

    def delete(self, hash: types.UInt256) -> None:
        """
        Remove a transaction.

        Args:
            hash: transaction hash.
        """
        super(CachedTXAccess, self)._delete(hash)

    def all(self) -> Iterator[payloads.Transaction]:
        """
        Retrieve all transactions (readonly)
        """
        transactions = []
        for tx in self._internal_all():
            if tx.hash() not in self._dictionary:
                transactions.append(tx)

        for k, v in self._dictionary.items():
            if v.state != TrackState.DELETED:
                transactions.append(deepcopy(v.item))

        transactions.sort(key=lambda tx: tx.hash().to_array())
        for tx in transactions:
            yield tx
        return None


"""
Yes there is near duplicate code for all the Clone_xxx_Cache classes. 

A solution with a generic CloneCache and multiple inheritance quickly became hard to follow in part because Python 
requires all parent classes to be designed for multiple inheritance (where our case needs different super args). The 
solution using **kwargs in combination with no manual controllable MRO becomes a mess. 

Revisit later when there's more time to think about the problem again. 
"""


class CloneBlockCache(CachedBlockAccess):
    def __init__(self, db, inner_cache: CachedBlockAccess):
        super(CloneBlockCache, self).__init__(db)
        self.inner_cache = inner_cache
        self._internal_get = self._inner_cache_get
        self._internal_try_get = self._inner_cache_try_get
        self._internal_all = self._inner_cache_all

    def _inner_cache_try_get(self, hash, read_only=True):
        try:
            return self._inner_cache_get(hash, read_only)
        except KeyError:
            return None

    def _inner_cache_get(self, hash, read_only=True):
        return self.inner_cache.get(hash, read_only)

    def _inner_cache_all(self):
        return self.inner_cache.all()

    def commit(self) -> None:
        """
        Persist changes to the parent snapshot.
        """
        for trackable in self._dictionary.values():  # trackable.item: payloads.Block
            if trackable.state == TrackState.ADDED:
                self.inner_cache.put(trackable.item)
            elif trackable.state == TrackState.CHANGED:
                # This one is kind of useless unless we augment Block with additional attributes (like done with
                # Transaction) and the IDBImplementation class calls are updated to serialize_special(). Otherwise, any
                # attribute changes will modify the hash of the Block causing it to not be found.
                item = self.inner_cache.try_get(trackable.item.hash(), read_only=False)
                if item:
                    item.from_replica(trackable.item)
            elif trackable.state == TrackState.DELETED:
                self.inner_cache.delete(trackable.item.hash())


class CloneContractCache(CachedContractAccess):
    def __init__(self, db, inner_cache: CachedContractAccess):
        super(CloneContractCache, self).__init__(db)
        self.inner_cache = inner_cache
        self._internal_get = self._inner_cache_get
        self._internal_try_get = self._inner_cache_try_get
        self._internal_all = self._inner_cache_all

    def _inner_cache_try_get(self, hash, read_only=True):
        try:
            return self._inner_cache_get(hash, read_only)
        except KeyError:
            return None

    def _inner_cache_get(self, hash, read_only=True):
        return self.inner_cache.get(hash, read_only)

    def _inner_cache_all(self):
        return self.inner_cache.all()

    def commit(self) -> None:
        """
        Persist changes to the parent snapshot.
        """
        for trackable in self._dictionary.values():  # trackable.item: storage.ContractState
            if trackable.state == TrackState.ADDED:
                self.inner_cache.put(trackable.item)
            elif trackable.state == TrackState.CHANGED:
                item = self.inner_cache.try_get(trackable.item.script_hash(), read_only=False)
                if item:
                    item.from_replica(trackable.item)
            elif trackable.state == TrackState.DELETED:
                self.inner_cache.delete(trackable.item.script_hash())


class CloneStorageCache(CachedStorageAccess):
    def __init__(self, db, inner_cache: CachedStorageAccess):
        super(CloneStorageCache, self).__init__(db)
        self.inner_cache = inner_cache
        self._internal_get = self._inner_cache_get
        self._internal_try_get = self._inner_cache_try_get
        self._internal_all = self._inner_cache_all
        self._internal_find = self._inner_cache_find

    def _inner_cache_try_get(self, hash, read_only=True):
        try:
            return self._inner_cache_get(hash, read_only)
        except KeyError:
            return None

    def _inner_cache_get(self, hash, read_only=True):
        return self.inner_cache.get(hash, read_only)

    def _inner_cache_all(self, contract_script_hash):
        return self.inner_cache.all(contract_script_hash)

    def _inner_cache_find(self, contract_script_hash, key_prefix):
        return self.inner_cache.find(contract_script_hash, key_prefix)

    def commit(self) -> None:
        """
        Persist changes to the parent snapshot.
        """
        for trackable in self._dictionary.values():
            if trackable.state == TrackState.ADDED:
                self.inner_cache.put(trackable.key, trackable.item)
            elif trackable.state == TrackState.CHANGED:
                item = self.inner_cache.try_get(trackable.key, read_only=False)
                if item:
                    item.from_replica(trackable.item)
            elif trackable.state == TrackState.DELETED:
                self.inner_cache.delete(trackable.key)


class CloneTXCache(CachedTXAccess):
    def __init__(self, db, inner_cache: CachedTXAccess):
        super(CloneTXCache, self).__init__(db)
        self.inner_cache = inner_cache
        self._internal_get = self._inner_cache_get
        self._internal_try_get = self._inner_cache_try_get
        self._internal_all = self._inner_cache_all

    def _inner_cache_try_get(self, hash, read_only=True):
        try:
            return self._inner_cache_get(hash, read_only)
        except KeyError:
            return None

    def _inner_cache_get(self, hash, read_only=True):
        return self.inner_cache.get(hash, read_only)

    def _inner_cache_all(self):
        return self.inner_cache.all()

    def commit(self) -> None:
        """
        Persist changes to the parent snapshot.
        """
        for trackable in self._dictionary.values():  # trackable.item: payloads.Transaction
            if trackable.state == TrackState.ADDED:
                self.inner_cache.put(trackable.item)
            elif trackable.state == TrackState.CHANGED:
                # Note that any changes to TX attributes that are serialized changes the hash and thus we won't find it
                # anymore in the cache.
                #
                # This means the changes will be persisted in the DB with a new entry!
                # We can however augment the TX class with e.g. VMState and BlockIndex without affecting the hash.
                # For all other cases the behaviour of a `transaction.get(hash)` is equal to
                # `transaction.get(hash, readonly=True) as no data will be persisted.
                item = self.inner_cache.try_get(trackable.item.hash(), read_only=False)
                if item:
                    item.from_replica(trackable.item)
            elif trackable.state == TrackState.DELETED:
                self.inner_cache.delete(trackable.item.hash())


class CloneAttributeCache(AttributeCache):
    def __init__(self, inner_cache: AttributeCache):
        super(CloneAttributeCache, self).__init__()
        self._inner_cache = inner_cache

    def _get_internal(self):
        return self._inner_cache.get()

    def _update_internal(self, value):
        self._inner_cache.put(value)



import os
from typing import Any, Dict, Optional, Tuple
from unittest import TestCase

from boa3.analyser.analyser import Analyser
from boa3.compiler.compiler import Compiler


class BoaTest(TestCase):
    dirname: str = None

    @classmethod
    def setUpClass(cls):
        path = os.path.abspath(__file__).replace('\\', '/')  # for windows compatibility
        cls.dirname = '/'.join(path.split('/')[:-3])

        super(BoaTest, cls).setUpClass()

    def get_compiler_analyser(self, compiler: Compiler) -> Analyser:
        return compiler._analyser

    def assertCompilerLogs(self, expected_logged_exception, path):
        output = None
        with self.assertLogs() as log:
            from boa3.exception.NotLoadedException import NotLoadedException
            try:
                from boa3.boa3 import Boa3
                output = Boa3.compile(path)
            except NotLoadedException:
                # when an compiler error is logged this exception is raised.
                pass

        for logger in log.records:
            import logging
            logging.log(level=logger.levelno, msg=logger.msg)

        if len([exception for exception in log.records if isinstance(exception.msg, expected_logged_exception)]) <= 0:
            raise AssertionError('{0} not logged'.format(expected_logged_exception.__name__))
        return output

    def compile_and_save(self, path: str) -> Tuple[bytes, Dict[str, Any]]:
        nef_output = path.replace('.py', '.nef')
        manifest_output = path.replace('.py', '.manifest.json')

        from boa3.boa3 import Boa3
        from boa3.neo.contracts.neffile import NefFile
        Boa3.compile_and_save(path)

        with open(nef_output, mode='rb') as nef:
            file = nef.read()
            output = NefFile.deserialize(file).script

        with open(manifest_output) as manifest_output:
            import json
            manifest = json.loads(manifest_output.read())

        return output, manifest

    def get_debug_info(self, path: str) -> Optional[Dict[str, Any]]:
        debug_info_output = path.replace('.py', '.nefdbgnfo')

        if not os.path.isfile(debug_info_output):
            return None

        from zipfile import ZipFile
        with ZipFile(debug_info_output, 'r') as dbgnfo:
            import json
            debug_info = json.loads(dbgnfo.read(os.path.basename(path.replace('.py', '.debug.json'))))
        return debug_info



from boa3.boa3 import Boa3
from boa3.builtin.interop.runtime import TriggerType
from boa3.exception.CompilerError import MismatchedTypes
from boa3.model.builtin.interop.interop import Interop
from boa3.model.type.type import Type
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest


class TestInterop(BoaTest):

    def test_check_witness(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.CheckWitness.interop_method_hash
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/interop_test/CheckWitness.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_check_witness_imported_as(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.CheckWitness.interop_method_hash
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/interop_test/CheckWitnessImportedAs.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_check_witness_mismatched_type(self):
        path = '%s/boa3_test/test_sc/interop_test/CheckWitnessMismatchedType.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_notify_str(self):
        event_name = String('notify').to_bytes()
        string = String('str').to_bytes()
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(string)).to_byte_array(min_length=1)
            + string
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/interop_test/NotifyStr.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_notify_int(self):
        event_name = String('notify').to_bytes()
        expected_output = (
            Opcode.PUSH15
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/interop_test/NotifyInt.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_notify_bool(self):
        event_name = String('notify').to_bytes()
        expected_output = (
            Opcode.PUSH1
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/interop_test/NotifyBool.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_notify_none(self):
        event_name = String('notify').to_bytes()
        expected_output = (
            Opcode.PUSHNULL
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/interop_test/NotifyNone.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_notify_sequence(self):
        event_name = String('notify').to_bytes()
        expected_output = (
            Opcode.PUSH7
            + Opcode.PUSH5
            + Opcode.PUSH3
            + Opcode.PUSH2
            + Opcode.PUSH4
            + Opcode.PACK
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/interop_test/NotifySequence.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_log_mismatched_type(self):
        path = '%s/boa3_test/test_sc/interop_test/LogMismatchedValueInt.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_log_str(self):
        string = String('str').to_bytes()
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(string)).to_byte_array(min_length=1)
            + string
            + Opcode.SYSCALL
            + Interop.Log.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/interop_test/LogStr.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_get_trigger(self):
        expected_output = (
            Opcode.SYSCALL
            + Interop.GetTrigger.interop_method_hash
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/interop_test/Trigger.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_is_application_trigger(self):
        application = Integer(TriggerType.APPLICATION.value).to_byte_array()
        expected_output = (
            Opcode.SYSCALL
            + Interop.GetTrigger.interop_method_hash
            + Opcode.PUSHDATA1
            + Integer(len(application)).to_byte_array(min_length=1)
            + application
            + Opcode.CONVERT
            + Type.int.stack_item
            + Opcode.EQUAL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/interop_test/TriggerApplication.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_is_verification_trigger(self):
        verification = Integer(TriggerType.VERIFICATION.value).to_byte_array()
        expected_output = (
            Opcode.SYSCALL
            + Interop.GetTrigger.interop_method_hash
            + Opcode.PUSHDATA1
            + Integer(len(verification)).to_byte_array(min_length=1)
            + verification
            + Opcode.CONVERT
            + Type.int.stack_item
            + Opcode.EQUAL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/interop_test/TriggerVerification.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_get_calling_script_hash(self):
        expected_output = (
            Opcode.SYSCALL
            + Interop.CallingScriptHash.getter.interop_method_hash
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/interop_test/CallingScriptHash.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)



from typing import Dict

from boa3.model.builtin.interop.interopevent import InteropEvent
from boa3.model.variable import Variable


class NotifyMethod(InteropEvent):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'notify'
        syscall = 'System.Runtime.Notify'
        args: Dict[str, Variable] = {'state': Variable(Type.any)}
        super().__init__(identifier, syscall, args)



from enum import Enum
from typing import List, Dict

from boa3.model.builtin.interop.runtime.checkwitnessmethod import CheckWitnessMethod
from boa3.model.builtin.interop.runtime.getcallingscripthashmethod import CallingScriptHashProperty
from boa3.model.builtin.interop.runtime.logmethod import LogMethod
from boa3.model.builtin.interop.runtime.notifymethod import NotifyMethod
from boa3.model.builtin.interop.runtime.triggermethod import TriggerMethod
from boa3.model.builtin.interop.runtime.triggertype import TriggerType as TriggerTyping
from boa3.model.builtin.interop.storage.storagedeletemethod import StorageDeleteMethod
from boa3.model.builtin.interop.storage.storagegetmethod import StorageGetMethod
from boa3.model.builtin.interop.storage.storageputmethod import StoragePutMethod
from boa3.model.identifiedsymbol import IdentifiedSymbol


class InteropPackage(str, Enum):
    Runtime = 'runtime'
    Storage = 'storage'


class Interop:

    @classmethod
    def interop_symbols(cls, package: str = None) -> List[IdentifiedSymbol]:
        if package in InteropPackage.__members__.values():
            return cls._interop_symbols[package]

        lst: List[IdentifiedSymbol] = []
        for symbols in cls._interop_symbols.values():
            lst.extend(symbols)
        return lst

    # Runtime Interops
    CheckWitness = CheckWitnessMethod()
    Notify = NotifyMethod()
    Log = LogMethod()
    TriggerType = TriggerTyping()
    GetTrigger = TriggerMethod(TriggerType)
    CallingScriptHash = CallingScriptHashProperty()

    # Storage Interops
    StorageGet = StorageGetMethod()
    StoragePut = StoragePutMethod()
    StorageDelete = StorageDeleteMethod()

    _interop_symbols: Dict[InteropPackage, List[IdentifiedSymbol]] = {
        InteropPackage.Runtime: [CheckWitness,
                                 Notify,
                                 Log,
                                 TriggerType,
                                 GetTrigger,
                                 CallingScriptHash
                                 ],
        InteropPackage.Storage: [StorageGet,
                                 StoragePut,
                                 StorageDelete
                                 ]
    }



from typing import Any, Set

from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.itype import IType


class GenericSequenceType(SequenceType):
    """
    An class used to represent a generic Python sequence type
    """

    def __init__(self, values_type: Set[IType] = None):
        identifier: str = 'sequence'
        values_type = self.filter_types(values_type)
        super().__init__(identifier, values_type)

    def is_valid_key(self, key_type: IType) -> bool:
        return key_type == self.valid_key

    @property
    def valid_key(self) -> IType:
        from boa3.model.type.type import Type
        return Type.int

    @property
    def is_generic(self) -> bool:
        return True

    @classmethod
    def build(cls, value: Any):
        if cls._is_type_of(value):
            values_types: Set[IType] = cls.get_types(value)
            return cls(values_types)

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, SequenceType)

    def __hash__(self):
        return hash(self.identifier)



from typing import Optional

from boa3.model.operation.operator import Operator
from boa3.model.operation.unary.booleannot import BooleanNot
from boa3.model.operation.unary.logicnot import LogicNot
from boa3.model.operation.unary.negative import Negative
from boa3.model.operation.unary.positive import Positive
from boa3.model.operation.unary.unaryoperation import UnaryOperation
from boa3.model.type.itype import IType


class UnaryOp:
    # Arithmetic operations
    Positive = Positive()
    Negative = Negative()

    # Logical operations
    Not = BooleanNot()
    BitNot = LogicNot()

    @classmethod
    def validate_type(cls, operator: Operator, operand: IType) -> Optional[UnaryOperation]:
        """
        Gets a unary operation given the operator and the operand type.

        :param operator: unary operator
        :param operand: type of the operand
        :return: The operation if exists. None otherwise;
        :rtype: UnaryOperation or None
        """
        for id, op in vars(cls).items():
            if isinstance(op, UnaryOperation) and op.is_valid(operator, operand):
                return op.build(operand)

    @classmethod
    def get_operation_by_operator(cls, operator: Operator) -> Optional[UnaryOperation]:
        """
        Gets a unary operation given the operator.

        :param operator: unary operator
        :return: The operation if exists. If exists more than one operation with the same operator, returns the first
        found. None otherwise.
        :rtype: UnaryOperation or None
        """
        for id, op in vars(cls).items():
            if isinstance(op, UnaryOperation) and op.operator is operator:
                return op

    @classmethod
    def get_operation(cls, operation: UnaryOperation) -> Optional[UnaryOperation]:
        """
        Gets an unary operation given another operation.

        :param operation: unary operation
        :return: The operation if exists. None otherwise;
        :rtype: UnaryOperation or None
        """
        for id, op in vars(cls).items():
            if type(operation) == type(op):
                return op



import boa3
from boa3.neo.contracts import NEF, Version
from boa3.neo.core import BinaryReader, BinaryWriter


class NefFile:
    """
    The object encapsulates the information of the NEO Executable Format (NEF)

    :ivar _nef: nef serializable object
    """

    def __init__(self, script_bytes: bytes):
        """
        :param script_bytes: the script of the smart contract
        """
        compiler: str = "neo3-boa by COZ"
        version = Version.from_string(boa3.__version__)
        self._nef = NEF(compiler, version, script_bytes)

    @property
    def script(self) -> bytes:
        return self._nef.script

    @property
    def script_hash(self) -> bytes:
        return self._nef.script_hash.to_array()

    def _set_version(self, version: str):
        """
        Updates the NEF file compiler version
        """
        try:
            version = Version.from_string(version)
            self._nef.version = version
        except Exception as e:
            import logging
            logging.error(str(e))

    def serialize(self) -> bytes:
        """
        Serialize the NefFile object

        :return: bytes of the serialized object.
        """
        with BinaryWriter() as writer:
            self._nef.serialize(writer)
            result = writer.to_array()
        return result

    @classmethod
    def deserialize(cls, bts: bytes):
        """
        Deserialize the NefFile object

        :param bts: stream data
        :return: the deserialized object.
        :rtype: NefFile
        """
        with BinaryReader(bts) as reader:
            nef = NEF()
            nef.deserialize(reader)

            nef_file = cls(b'')
            nef_file._nef = nef
        return nef_file



from typing import Any


def get(key: Any) -> bytes:
    """
    Gets a value from the persistent store based on the given key.

    :param key: value identifier in the store
    :type key: str or bytes
    :return: the value corresponding to given key for current storage context
    :rtype: bytes
    """
    pass


def put(key: Any, value: Any):
    """
    Inserts a given value in the key-value format into the persistent storage.

    :param key: the identifier in the store for the new value
    :type key: str or bytes
    :param value: value to be stored
    :type value: int or str or bytes
    """
    pass


def delete(key: Any):
    """
    Removes a given key from the persistent storage if exists.

    :param key: the identifier in the store for the new value
    :type key: str or bytes
    """
    pass



from __future__ import annotations

import hashlib
from copy import deepcopy
from typing import List

from bitarray import bitarray  # type: ignore

from boa3.neo3.core import Size as s, serialization, types, utils, cryptography as crypto, IClonable
from boa3.neo3.network import payloads


class _BlockBase(serialization.ISerializable):
    def __init__(self,
                 version: int = 0,
                 prev_hash: types.UInt256 = None,
                 merkle_root: types.UInt256 = None,
                 timestamp: int = 0,
                 index: int = 0,
                 next_consensus: types.UInt160 = None,
                 witness: payloads.Witness = None
                 ):

        self.version = version
        self.prev_hash = prev_hash if prev_hash else types.UInt256.zero()
        self.merkle_root = merkle_root if merkle_root else types.UInt256.zero()
        self.timestamp = timestamp
        self.index = index
        self.next_consensus = next_consensus if next_consensus else types.UInt160.zero()
        self.witness = witness if witness else payloads.Witness()

    def __len__(self):
        return s.uint32 + s.uint256 + s.uint256 + s.uint64 + s.uint32 + s.uint160 + 1 + len(self.witness)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        self.serialize_unsigned(writer)
        writer.write_uint8(1)
        writer.write_serializable(self.witness)

    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        writer.write_uint32(self.version)
        writer.write_serializable(self.prev_hash)
        writer.write_serializable(self.merkle_root)
        writer.write_uint64(self.timestamp)
        writer.write_uint32(self.index)
        writer.write_serializable(self.next_consensus)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if no witnesses are found.
        """
        self.deserialize_unsigned(reader)
        witness_obj_count = reader.read_uint8()
        if witness_obj_count != 1:
            raise ValueError(f"Deserialization error - Witness object count is {witness_obj_count} must be 1")
        self.witness = reader.read_serializable(payloads.Witness)

    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        self.version = reader.read_uint32()
        self.prev_hash = reader.read_serializable(types.UInt256)
        self.merkle_root = reader.read_serializable(types.UInt256)
        self.timestamp = reader.read_uint64()
        self.index = reader.read_uint32()
        self.next_consensus = reader.read_serializable(types.UInt160)

    def hash(self) -> types.UInt256:
        """
        Get a unique block identifier based on the unsigned data portion of the object.
        """
        with serialization.BinaryWriter() as bw:
            self.serialize_unsigned(bw)
            data_to_hash = bytearray(bw._stream.getvalue())
            data = hashlib.sha256(hashlib.sha256(data_to_hash).digest()).digest()
            return types.UInt256(data=data)


class Header(_BlockBase):
    """
    A Block header only object.

    Does not contain any consensus data or transactions.

    See also:
        :class:`~neo3.network.payloads.block.TrimmedBlock`
    """

    def __init__(self,
                 version: int = 0,
                 prev_hash: types.UInt256 = None,
                 merkle_root: types.UInt256 = None,
                 timestamp: int = 0,
                 index: int = 0,
                 next_consensus: types.UInt160 = None,
                 witness: payloads.Witness = None
                 ):
        super(Header, self).__init__(version, prev_hash, merkle_root, timestamp, index, next_consensus, witness)

    def __len__(self):
        return super(Header, self).__len__() + 1

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.hash() != other.hash():
            return False
        return True

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        super(Header, self).serialize(writer)
        writer.write_uint8(0)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the check byte does not equal.
        """
        super(Header, self).deserialize(reader)
        tmp = reader.read_uint8()
        if tmp != 0:
            raise ValueError("Deserialization error")


class Block(_BlockBase, payloads.IInventory):
    """
    The famous Block. I transfer chain state.
    """
    #: The maximum item count per block. Consensus data and Transactions are considered items.
    MAX_CONTENTS_PER_BLOCK = 65535
    #: The maximum number of transactions allowed to be included in a block.
    MAX_TX_PER_BLOCK = MAX_CONTENTS_PER_BLOCK - 1

    def __init__(self,
                 version: int = 0,
                 prev_hash: types.UInt256 = None,
                 merkle_root: types.UInt256 = None,
                 timestamp: int = 0,
                 index: int = 0,
                 next_consensus: types.UInt160 = None,
                 witness: payloads.Witness = None,
                 consensus_data: payloads.ConsensusData = None,
                 transactions: List[payloads.Transaction] = None
                 ):
        super(Block, self).__init__(version, prev_hash, merkle_root, timestamp, index, next_consensus, witness)
        self.consensus_data = consensus_data if consensus_data else payloads.ConsensusData()
        self.transactions = [] if transactions is None else transactions

    def __len__(self):
        # calculate the varint length that needs to be inserted before the transaction objects.
        magic_len = utils.get_var_size(len(self.transactions))
        txs_len = sum([len(t) for t in self.transactions])
        return super(Block, self).__len__() + magic_len + len(self.consensus_data) + txs_len

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.hash() != other.hash():
            return False
        return True

    @property
    def inventory_type(self) -> payloads.InventoryType:
        """
        Inventory type identifier.
        """
        return payloads.InventoryType.BLOCK

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        super(Block, self).serialize(writer)
        writer.write_var_int(len(self.transactions) + 1)
        writer.write_serializable(self.consensus_data)
        for tx in self.transactions:
            writer.write_serializable(tx)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the content count of the block is zero, or if there is a duplicate transaction in the list,
                or if the merkle root does not included the calculated root.
        """
        super(Block, self).deserialize(reader)
        content_count = reader.read_var_int(max=self.MAX_CONTENTS_PER_BLOCK)
        if content_count == 0:
            raise ValueError("Deserialization error - no contents")

        self.consensus_data = reader.read_serializable(payloads.ConsensusData)
        tx_count = content_count - 1
        for _ in range(tx_count):
            self.transactions.append(reader.read_serializable(payloads.Transaction))

        if len(set(self.transactions)) != tx_count:
            raise ValueError("Deserialization error - block contains duplicate transaction")

        hashes = [t.hash() for t in self.transactions]
        if Block.calculate_merkle_root(self.consensus_data.hash(), hashes) != self.merkle_root:
            raise ValueError("Deserialization error - merkle root mismatch")

    def rebuild_merkle_root(self) -> None:
        """
        Recalculates the Merkle root.
        """
        self.merkle_root = Block.calculate_merkle_root(self.consensus_data.hash(),
                                                       [t.hash() for t in self.transactions])

    def trim(self) -> TrimmedBlock:
        """
        Reduce a block in size by replacing the consensus data and transaction objects with their identifying hashes.
        """
        hashes = [self.consensus_data.hash()] + [t.hash() for t in self.transactions]
        return TrimmedBlock(version=self.version,
                            prev_hash=self.prev_hash,
                            merkle_root=self.merkle_root,
                            timestamp=self.timestamp,
                            index=self.index,
                            next_consensus=self.next_consensus,
                            witness=self.witness,
                            hashes=hashes,
                            consensus_data=self.consensus_data
                            )

    @staticmethod
    def calculate_merkle_root(consensus_data_hash: types.UInt256,
                              transaction_hashes: List[types.UInt256]) -> types.UInt256:
        """
        Calculate a Merkle root.

        Args:
            consensus_data_hash:
            transaction_hashes:
        """
        hashes = [consensus_data_hash] + transaction_hashes
        return crypto.MerkleTree.compute_root(hashes)

    def from_replica(self, replica: Block) -> None:
        self.version = replica.version
        self.prev_hash = replica.prev_hash
        self.merkle_root = replica.merkle_root
        self.timestamp = replica.timestamp
        self.index = replica.index
        self.next_consensus = replica.next_consensus
        self.witness = replica.witness
        self.consensus_data = replica.consensus_data
        self.transactions = replica.transactions


class TrimmedBlock(_BlockBase, IClonable):
    """
    A size reduced Block instance.

    Contains consensus data and transactions hashes instead of their full objects.
    """

    def __init__(self,
                 version: int = 0,
                 prev_hash: types.UInt256 = None,
                 merkle_root: types.UInt256 = None,
                 timestamp: int = 0,
                 index: int = 0,
                 next_consensus: types.UInt160 = None,
                 witness: payloads.Witness = None,
                 hashes: List[types.UInt256] = None,
                 consensus_data: payloads.ConsensusData = None):
        super(TrimmedBlock, self).__init__(version, prev_hash, merkle_root, timestamp, index, next_consensus, witness)
        self.hashes = hashes if hashes else []
        self.consensus_data = consensus_data if consensus_data else payloads.ConsensusData()

    def __len__(self):
        size = super(TrimmedBlock, self).__len__()
        size += utils.get_var_size(self.hashes)
        if self.consensus_data:
            size += len(self.consensus_data)
        return size

    def __deepcopy__(self, memodict={}):
        # not the best, but faster than letting deepcopy() do introspection
        return TrimmedBlock.deserialize_from_bytes(self.to_array())

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        super(TrimmedBlock, self).serialize(writer)
        writer.write_serializable_list(self.hashes)
        if len(self.hashes) > 0:
            writer.write_serializable(self.consensus_data)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        super(TrimmedBlock, self).deserialize(reader)
        self.hashes = reader.read_serializable_list(types.UInt256)
        if len(self.hashes) > 0:
            self.consensus_data = reader.read_serializable(payloads.ConsensusData)

    def from_replica(self, replica: TrimmedBlock) -> None:
        """
        Shallow copy attributes from a reference object.
        """
        super().from_replica(replica)
        self.version = replica.version
        self.prev_hash = replica.prev_hash
        self.merkle_root = replica.merkle_root
        self.timestamp = replica.timestamp
        self.index = replica.index
        self.next_consensus = replica.next_consensus
        self.witness = replica.witness
        self.hashes = replica.hashes
        self.consensus_data = replica.consensus_data

    def clone(self) -> TrimmedBlock:
        """
        Deep copy
        """
        return deepcopy(self)


class MerkleBlockPayload(_BlockBase):
    def __init__(self,
                 version: int = 0,
                 prev_hash: types.UInt256 = None,
                 merkle_root: types.UInt256 = None,
                 timestamp: int = 0,
                 index: int = 0,
                 next_consensus: types.UInt160 = None,
                 witness: payloads.Witness = None,
                 content_count: int = None,
                 hashes: List[types.UInt256] = None,
                 flags: bytearray = None):
        super(MerkleBlockPayload, self).__init__(version, prev_hash, merkle_root,
                                                 timestamp, index, next_consensus, witness)
        """
        Should not be called directly. Use create() instead.
        """
        self.hashes = hashes if hashes else []
        self.content_count = content_count if content_count else len(self.hashes)
        self.flags = flags if flags else b''

    def __len__(self):
        return super(MerkleBlockPayload, self).__len__() + s.uint32 + utils.get_var_size(self.hashes) + \
            utils.get_var_size(self.flags)

    @classmethod
    def create(cls, block: Block, flags: bitarray) -> MerkleBlockPayload:
        """
        Create payload.

        Should be used instead of directly calling the initializer.

        Args:
            block:
            flags:
        """
        hashes = [block.consensus_data.hash()] + [t.hash() for t in block.transactions]
        tree = crypto.MerkleTree(hashes)
        flag_bytes = flags.tobytes()

        return cls(version=block.version, prev_hash=block.prev_hash, merkle_root=block.merkle_root,
                   timestamp=block.timestamp, index=block.index, next_consensus=block.next_consensus,
                   witness=block.witness, content_count=len(hashes), hashes=tree.to_hash_array(),
                   flags=flag_bytes)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        super(MerkleBlockPayload, self).serialize(writer)
        writer.write_var_int(self.content_count)
        writer.write_serializable_list(self.hashes)
        writer.write_var_bytes(self.flags)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        super(MerkleBlockPayload, self).deserialize(reader)
        self.content_count = reader.read_var_int()
        self.hashes = reader.read_serializable_list(types.UInt256)
        self.flags = reader.read_var_bytes()


class HeadersPayload(serialization.ISerializable):
    MAX_HEADERS_COUNT = 2000

    def __init__(self, headers: List[Header] = None):
        """
        Should not be called directly. Use create() instead.
        """
        self.headers = headers if headers else []

    def __len__(self):
        return utils.get_var_size(self.headers)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable_list(self.headers)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.headers = reader.read_serializable_list(Header)

    @classmethod
    def create(cls, headers: List[Header]) -> HeadersPayload:
        """
        Create payload.

        Args:
            headers: Header objects to include.
        """
        return cls(headers)


class GetBlocksPayload(serialization.ISerializable):
    """
    Used to request an array block hashes that can be retrieved via a message with the
    :const:`~neo3.network.message.MessageType.GETDATA` type.
    """

    def __init__(self, hash_start: types.UInt256 = None, count=-1):
        """
        Should not be called directly. Use create() instead.
        """
        self.hash_start = hash_start if hash_start else types.UInt256.zero()
        self.count = count

    def __len__(self):
        return s.uint256 + s.uint16

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.hash_start)
        writer.write_int16(self.count)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.hash_start = reader.read_serializable(types.UInt256)
        self.count = reader.read_int16()

    @classmethod
    def create(cls, hash_start: types.UInt256, count=-1) -> GetBlocksPayload:
        """
        Create payload.

        Args:
            hash_start: starting point from which to return the `next` hash.

                Note:

                    For syncing supply the local best height block hash to receive the hashes in the range of
                    best_height+1 to best_height+1+count

            count: number of hashes to return.
        """
        return cls(hash_start, count)


class GetBlockDataPayload(serialization.ISerializable):
    """
    Used to request full Block objects via a message with the :const:`~neo3.network.message.MessageType.GETBLOCKDATA`
    type.
    """
    MAX_BLOCKS_COUNT = 500

    def __init__(self, index_start: int = 0, count: int = 500):
        """
        Should not be called directly. Use create() instead.
        """
        self.index_start = index_start
        self.count = count

    def __len__(self):
        return s.uint32 + s.uint16

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint32(self.index_start)
        writer.write_uint16(self.count)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if `count` is zero or exceeds
               :const:`~neo3.network.payloads.getblocks.GetBlockDataPayload.MAX_BLOCKS_COUNT`.
        """
        self.index_start = reader.read_uint32()
        self.count = reader.read_uint16()
        if self.count == 0 or self.count > self.MAX_BLOCKS_COUNT:
            raise ValueError("Deserialization error - invalid count")

    @classmethod
    def create(cls, index_start: int, count: int = 500) -> GetBlockDataPayload:
        """
        Create payload.

        Args:
            index_start: start block height.
            count: number of blocks to requests starting from `index_start`.
        """



from __future__ import annotations

import hashlib
from enum import IntEnum
from typing import TypeVar

from boa3.neo3.core import Size as s, serialization, utils, types
from boa3.neo3.network import payloads

ConsensusMessage_t = TypeVar('ConsensusMessage_t', bound='ConsensusMessage')


class ConsensusMessageType(IntEnum):
    CHANGE_VIEW = 0x00,
    PREPARE_REQUEST = 0x20,
    PREPARE_RESPONSE = 0x21,
    COMMIT = 0x30,
    RECOVERY_REQUEST = 0x40,
    RECOVERY_MESSAGE = 0x41


class ConsensusMessage(serialization.ISerializable):
    """
    Base class for the various consensus messages
    """

    def __init__(self, type: ConsensusMessageType = None):
        self.type = type if type else ConsensusMessageType.CHANGE_VIEW
        self.view_number: int = 0

    def __len__(self):
        return s.uint8 + s.uint8

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint8(self.type)
        writer.write_uint8(self.view_number)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.type = ConsensusMessageType(reader.read_uint8())
        self.view_number = reader.read_uint8()

    def deserialize_specialization_from_bytes(self, data: bytearray) -> ConsensusMessage_t:
        # TODO: not implemented. Requires all ConsensusMessage subclasses to be implemented. Low priority.
        pass


class ConsensusPayload(serialization.ISerializable, payloads.IInventory):
    def __init__(self, version: int = 0,
                 prev_hash: types.UInt256 = None,
                 block_index: int = 0,
                 validator_index: int = 0,
                 data: bytes = None,
                 witness: payloads.Witness = None):
        self.version = version
        self.prev_hash = prev_hash if prev_hash else types.UInt256.zero()
        self.block_index = block_index
        self.validator_index = validator_index
        self.data = data if data else b''
        self.witness = witness if witness else payloads.Witness()

    def __len__(self):
        return (s.uint32 + len(self.prev_hash) + s.uint32 + s.uint16 + utils.get_var_size(self.data) + 1
                + len(self.witness))

    def hash(self) -> types.UInt256:
        with serialization.BinaryWriter() as bw:
            self.serialize_unsigned(bw)
            data_to_hash = bytearray(bw._stream.getvalue())
            data = hashlib.sha256(hashlib.sha256(data_to_hash).digest()).digest()
            return types.UInt256(data=data)

    @property
    def inventory_type(self):
        return payloads.InventoryType.CONSENSUS

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        self.serialize_unsigned(writer)
        writer.write_uint8(1)
        writer.write_serializable(self.witness)

    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream excluding the validation byte + witness.

        Args:
            writer: instance.
        """
        writer.write_uint32(self.version)
        writer.write_serializable(self.prev_hash)
        writer.write_uint32(self.block_index)
        writer.write_uint16(self.validator_index)
        writer.write_var_bytes(self.data)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the validation byte is not 1
        """
        self.deserialize_unsigned(reader)
        if reader.read_uint8() != 1:
            raise ValueError("Deserialization error - validation byte not 1")
        self.witness = reader.read_serializable(payloads.Witness)

    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream excluding the validation byte + witness.

        Args:
            reader: instance.
        """
        self.version = reader.read_uint32()
        self.prev_hash = reader.read_serializable(types.UInt256)
        self.block_index = reader.read_uint32()
        self.validator_index = reader.read_uint16()
        self.data = reader.read_var_bytes()

    @classmethod
    def create(cls, version: int, prev_hash: types.UInt256, block_index: int, validator_index: int, data: bytes,
               witness: payloads.Witness) -> ConsensusPayload:
        return cls(version, prev_hash, block_index, validator_index, data, witness)


class ConsensusData(serialization.ISerializable):
    def __init__(self, primary_index: int = 0, nonce: int = 0):
        self.primary_index = primary_index
        self.nonce = nonce

    def __len__(self):
        return utils.get_var_size(self.primary_index) + s.uint64

    def hash(self) -> types.UInt256:
        data = hashlib.sha256(hashlib.sha256(self.to_array()).digest()).digest()
        return types.UInt256(data=data)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_var_int(self.primary_index)
        writer.write_uint64(self.nonce)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.primary_index = reader.read_var_int(max=1024)  # comes from C#'s Clockchain.MaxValidators
        self.nonce = reader.read_uint64()



from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MismatchedTypes
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3_test.tests.boa_test import BoaTest


class TestAny(BoaTest):

    def test_variable_none(self):
        path = '%s/boa3_test/test_sc/none_test/VariableNone.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHNULL
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_none_tuple(self):
        path = '%s/boa3_test/test_sc/none_test/NoneTuple.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHNULL   # a = (None, None, None)
            + Opcode.PUSHNULL
            + Opcode.PUSHNULL
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_none_identity(self):
        path = '%s/boa3_test/test_sc/none_test/NoneIdentity.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.ISNULL
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_none_not_identity(self):
        path = '%s/boa3_test/test_sc/none_test/NoneNotIdentity.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.ISNULL
            + Opcode.NOT
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_none_equality(self):
        path = '%s/boa3_test/test_sc/none_test/NoneEquality.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_mismatched_type_int_operation(self):
        path = '%s/boa3_test/test_sc/none_test/MismatchedTypesInOperation.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_mismatched_type_assign(self):
        path = '%s/boa3_test/test_sc/none_test/MismatchedTypesAssign.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_mismatched_type_after_reassign(self):
        path = '%s/boa3_test/test_sc/none_test/MismatchedTypesAfterReassign.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)



import asyncio
import struct
import traceback
import weakref
from asyncio import events
from asyncio.streams import StreamReader, StreamReaderProtocol, StreamWriter
from typing import Optional

from boa3.neo3 import network_logger as logger
from boa3.neo3.core import serialization
from boa3.neo3.network import node
from boa3.neo3.network.message import Message


class NeoProtocol(StreamReaderProtocol):
    def __init__(self, *args, **kwargs):
        """

        Args:
            *args:
            **kwargs:
        """
        sr = StreamReader()
        self._stream_reader_orig = sr
        self._stream_reader_wr = weakref.ref(sr)
        self._stream_writer = None
        self.client = node.NeoNode(self)
        self._loop = events.get_event_loop()
        super().__init__(sr)

    def connection_made(self, transport: asyncio.transports.BaseTransport) -> None:
        super().connection_made(transport)
        self._stream_writer = StreamWriter(transport, self, self._stream_reader_orig, self._loop)

        if self.client:
            asyncio.create_task(self.client.connection_made(transport))

    def connection_lost(self, exc: Optional[Exception] = None) -> None:
        if self.client:
            task = asyncio.create_task(self.client.connection_lost(exc))
            task.add_done_callback(lambda args: super(NeoProtocol, self).connection_lost(exc))
        else:
            super().connection_lost(exc)

    async def send_message(self, message: Message) -> None:
        try:
            self._stream_writer.write(message.to_array())
            await self._stream_writer.drain()
        except ConnectionResetError:
            print("connection reset")
            self.connection_lost(ConnectionResetError())
        except ConnectionError:
            print("connection error")
            self.connection_lost(ConnectionError())
        except asyncio.CancelledError:
            print("task cancelled, closing connection")
            # mypy can't seem to deduce that CancelledError still derives from Exception
            self.connection_lost(asyncio.CancelledError())  # type: ignore
        except Exception as e:
            print(f"***** woah what happened here?! {traceback.format_exc()}")
            self.connection_lost(Exception())

    async def read_message(self, timeout: Optional[int] = 30) -> Optional[Message]:
        if timeout == 0:
            # avoid memleak. See: https://bugs.python.org/issue37042
            timeout = None

        async def _read():
            try:
                # readexactly can throw ConnectionResetError
                message_header = await self._stream_reader_orig.readexactly(3)
                payload_length = message_header[2]

                if payload_length == 0xFD:
                    len_bytes = await self._stream_reader_orig.readexactly(2)
                    payload_length, = struct.unpack("<H", len_bytes)
                elif payload_length == 0xFE:
                    len_bytes = await self._stream_reader_orig.readexactly(4)
                    payload_length, = struct.unpack("<I", len_bytes)
                elif payload_length == 0xFE:
                    len_bytes = await self._stream_reader_orig.readexactly(8)
                    payload_length, = struct.unpack("<Q", len_bytes)
                else:
                    len_bytes = b''

                if payload_length > Message.PAYLOAD_MAX_SIZE:
                    raise ValueError("Invalid format")

                payload_data = await self._stream_reader_orig.readexactly(payload_length)
                raw = message_header + len_bytes + payload_data

                with serialization.BinaryReader(raw) as br:
                    m = Message()
                    try:
                        m.deserialize(br)
                        return m
                    except Exception:
                        logger.debug(f"Failed to deserialize message: {traceback.format_exc()}")
                        return None

            except (ConnectionResetError, ValueError) as e:
                # ensures we break out of the main run() loop of Node, which triggers a disconnect callback to clean up
                self.client.disconnecting = True
                logger.debug(f"Failed to read message data for reason: {traceback.format_exc()}")
                return None
            except (asyncio.CancelledError, asyncio.IncompleteReadError):
                return None
            except Exception:
                # ensures we break out of the main run() loop of Node, which triggers a disconnect callback to clean up
                logger.debug(f"error read message 1 {traceback.format_exc()}")
                return None
        try:
            # logger.debug("trying to read message")
            return await asyncio.wait_for(_read(), timeout)
        except (asyncio.TimeoutError, asyncio.CancelledError):
            return None
        except Exception:
            logger.debug("error read message 2")
            traceback.print_exc()
            return None

    def disconnect(self) -> None:
        if self._stream_writer:
            self._stream_writer.close()



from __future__ import annotations

from boa3.neo3 import storage, settings
from boa3.neo3.network import payloads, convenience


class Blockchain(convenience._Singleton):

    def init(self, backend: storage.IDBImplementation = None):  # type: ignore
        self.backend = backend if backend else settings.database
        self._current_snapshot = None

    @property
    def currentSnapshot(self) -> storage.Snapshot:
        if self._current_snapshot is None:
            self._current_snapshot = self.backend.get_snapshotview()
        return self._current_snapshot  # type: ignore
        # mypy thinks it returns None

    @property
    def height(self):
        return self.currentSnapshot.block_height

    async def persist(self, block: payloads.Block):
        with self.backend.get_snapshotview() as snapshot:
            snapshot.block_height = block.index
            snapshot.blocks.put(block)

            for tx in block.transactions:
                tx.block_height = block.index
                snapshot.transactions.put(tx)

                # TODO: run VM

            snapshot.commit()
            self._current_snapshot = snapshot



from typing import Dict

from boa3.model.callable import Callable
from boa3.model.method import Method
from boa3.model.symbol import ISymbol
from boa3.model.variable import Variable


class Module(ISymbol):
    """
    A class used to represent a Python module

    :ivar variables: a dictionary that maps each variable with its name. Empty by default.
    :ivar methods: a dictionary that maps each method with its name. Empty by default.
    :ivar callables: a dictionary that maps each callable object with its name. Empty by default.
    :ivar imported_symbols: a dictionary that maps each imported symbol with its name. Empty by default.
    """

    def __init__(self, variables: Dict[str, Variable] = None, methods: Dict[str, Method] = None):
        if variables is None:
            variables = {}
        self.variables = variables

        if methods is None:
            methods = {}
        self.methods = methods
        self.callables: Dict[str, Callable] = {}

        self.imported_symbols = {}
        self.assigned_variables = []

    @property
    def shadowing_name(self) -> str:
        return 'module'

    def include_variable(self, var_id: str, var: Variable):
        """
        Includes a variable into the scope of the module

        :param var_id: variable identifier
        :param var: variable to be included
        """
        if var_id not in self.symbols:
            self.variables[var_id] = var

    def is_variable_assigned(self, var_id: str) -> bool:
        if var_id not in self.variables:
            return False

        if var_id in self.assigned_variables or var_id in self.imported_symbols:
            return True

        for imported in self.imported_symbols.values():
            from boa3.model.importsymbol import Import
            if isinstance(imported, Import) and self.variables[var_id] in imported.variables.values():
                return True

        return False

    def assign_variable(self, var_id: str):
        if var_id in self.variables:
            self.assigned_variables.append(var_id)

    def include_callable(self, method_id: str, method: Callable):
        """
        Includes a method into the scope of the module

        :param method_id: method identifier
        :param method: method to be included
        """
        if method_id not in self.symbols:
            if isinstance(method, Method):
                self.methods[method_id] = method
            else:
                self.callables[method_id] = method

    def include_symbol(self, symbol_id: str, symbol: ISymbol):
        """
        Includes a method into the scope of the module

        :param symbol_id: method identifier
        :param symbol: method to be included
        """
        if symbol_id not in self.symbols:
            if isinstance(symbol, Variable):
                self.include_variable(symbol_id, symbol)
            elif isinstance(symbol, Callable):
                self.include_callable(symbol_id, symbol)
            else:
                self.imported_symbols[symbol_id] = symbol

    @property
    def symbols(self) -> Dict[str, ISymbol]:
        """
        Gets all the symbols in the module

        :return: a dictionary that maps each symbol in the module with its name
        """
        symbols = {}
        symbols.update(self.imported_symbols)
        symbols.update(self.variables)
        symbols.update(self.methods)
        symbols.update(self.callables)
        return symbols



from .address import NetworkAddress, AddrPayload, AddressState, DisconnectReason
from .block import Header, Block, MerkleBlockPayload, HeadersPayload, TrimmedBlock, GetBlocksPayload, \
    GetBlockDataPayload
from .consensus import ConsensusData, ConsensusPayload
from .empty import EmptyPayload
from .filter import FilterAddPayload, FilterLoadPayload
from .inventory import InventoryPayload, InventoryType, IInventory
from .ping import PingPayload
from .transaction import Transaction, TransactionAttributeUsage, TransactionAttribute
from .verification import Witness, WitnessScope, Cosigner
from .version import VersionPayload

__all__ = ['EmptyPayload', 'InventoryPayload', 'InventoryType', 'VersionPayload', 'NetworkAddress',
           'AddrPayload', 'PingPayload', 'Witness', 'WitnessScope', 'Header', 'Block', 'MerkleBlockPayload',
           'HeadersPayload', 'ConsensusData', 'ConsensusPayload', 'Transaction', 'TransactionAttribute',
           'TransactionAttributeUsage', 'Cosigner', 'GetBlocksPayload', 'GetBlockDataPayload', 'FilterAddPayload',
           'FilterLoadPayload', 'TrimmedBlock']



def to_script_hash(data_bytes: bytes) -> bytes:
    """
    Converts a data to a script hash.

    :param data_bytes: data to hash.
    :type data_bytes: bytearray or bytes

    :return: the scripthash of the data
    :rtype: bytes
    """
    from boa3.neo import cryptography
    from base58 import b58decode
    try:
        base58_decoded = b58decode(data_bytes)[1:]  # first byte is the address version

        from boa3.constants import SIZE_OF_INT160
        if len(base58_decoded) < SIZE_OF_INT160:
            raise AttributeError
        return bytes(base58_decoded[:SIZE_OF_INT160])
    except BaseException:
        return cryptography.hash160(data_bytes)


def to_hex_str(data_bytes: bytes) -> str:
    """
    Converts bytes into its string hex representation.

    :param data_bytes: data to represent as hex.
    :type data_bytes: bytearray or bytes

    :return: the hex representation of the data
    :rtype: str
    """
    if isinstance(data_bytes, bytes):
        data_bytes = bytearray(data_bytes)
    data_bytes.reverse()
    return data_bytes.hex()



from abc import ABC, abstractmethod
from typing import List, Tuple

from boa3.model.operation.operator import Operator
from boa3.model.type.itype import IType
from boa3.neo.vm.opcode.Opcode import Opcode


class IOperation(ABC):
    """
    An interface used to represent operations

    :ivar operator: the operator of the operation
    :ivar result: the result type of the operation
    """

    def __init__(self, operator: Operator, result_type: IType):
        self.operator: Operator = operator
        self.result: IType = result_type

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        """
        Gets the operation sequence of opcodes with in Neo Vm

        :return: the opcode if exists. Empty list otherwise.
        """
        return []

    @property
    def bytecode(self) -> bytes:
        """
        Gets the operation bytecode

        :return: the bytecode if exists. Empty bytes otherwise.
        """
        str_mult_bytes = bytearray()
        for opcode, data in self.opcode:
            str_mult_bytes += opcode + data
        return bytes(str_mult_bytes)

    @property
    @abstractmethod
    def number_of_operands(self) -> int:
        """
        Gets the number of operands required for this operations

        :return: Number of operands
        """
        pass

    @property
    def op_on_stack(self) -> int:
        """
        Gets the number of arguments that must be on stack before the opcode is called.

        :return: the number of arguments. Same from `number_of_operands` by default.
        """
        return self.number_of_operands

    @abstractmethod
    def validate_type(self, *types: IType) -> bool:
        """
        Verifies if the given operands are valid to the operation

        :param types: types of the operand
        :return: True if all arguments are valid. False otherwise.
        """
        pass

    def is_valid(self, operator: Operator, *types: IType) -> bool:
        """
        Verifies if the given operator and operands are valid to the operation

        :param operator:
        :param types: types of the operand
        :return: True if all arguments are valid. False otherwise.
        """
        if len(types) != self.number_of_operands:
            return False

        return operator is self.operator and self.validate_type(*types)

    @property
    def is_supported(self) -> bool:
        """
        Verifies if the operation is supported by the compiler

        :return: True if it is supported. False otherwise.
        """
        return True

    @classmethod
    @abstractmethod
    def build(cls, *operands: IType):
        """
        Creates an operation with the given operands types

        :param operands: operands types
        :return: The built operation if the operands are valid. None otherwise
        :rtype: IOperation or None
        """
        return None



import ast
from typing import Dict

from boa3.analyser.constructanalyser import ConstructAnalyser
from boa3.analyser.moduleanalyser import ModuleAnalyser
from boa3.analyser.typeanalyser import TypeAnalyser
from boa3.builtin import NeoMetadata
from boa3.model.symbol import ISymbol
from boa3.model.type.type import Type


class Analyser(object):
    """
    This class is responsible for the semantic analysis of the code

    :ivar symbol_table: a dictionary used to store the identifiers
    """

    def __init__(self, ast_tree: ast.AST, path: str = None, log: bool = False):
        self.symbol_table: Dict[str, ISymbol] = {}

        self.ast_tree: ast.AST = ast_tree
        self.metadata: NeoMetadata = NeoMetadata()
        self.is_analysed: bool = False
        self._log: bool = log

        self.__include_builtins_symbols()

        import os
        self.path: str = path
        self.filename: str = path if path is None else os.path.split(os.path.realpath(path))[1]

    @staticmethod
    def analyse(path: str, log: bool = False):
        """
        Analyses the syntax of the Python code

        :param path: the path of the Python file
        :param log: if compiler errors should be logged.
        :return: a boolean value that represents if the analysis was successful
        :rtype: Analyser
        """
        with open(path, 'rb') as source:
            ast_tree = ast.parse(source.read())

        analyser = Analyser(ast_tree, path, log)
        analyser.__pre_execute()
        # fill symbol table
        if not analyser.__analyse_modules():
            return analyser
        # check is the types are correct
        if not analyser.__check_types():
            return analyser
        analyser.is_analysed = True

        return analyser

    def __include_builtins_symbols(self):
        """
        Include the Python builtins in the global symbol table
        """
        self.symbol_table.update(Type.builtin_types())

    def __check_types(self) -> bool:
        """
        Performs the type checking

        :return: a boolean value that represents if the analysis was successful
        """
        type_analyser = TypeAnalyser(self, self.symbol_table, log=self._log)
        return not type_analyser.has_errors

    def __analyse_modules(self) -> bool:
        """
        Validates the symbols and constructs the symbol table of the ast tree

        :return: a boolean value that represents if the analysis was successful
        """
        module_analyser = ModuleAnalyser(self, self.symbol_table, log=self._log)
        self.symbol_table.update(module_analyser.global_symbols)
        self.ast_tree.body.extend(module_analyser.imported_nodes)
        return not module_analyser.has_errors

    def __pre_execute(self):
        """
        Pre executes the instructions of the ast for optimization
        """
        self.ast_tree = ConstructAnalyser(self.ast_tree, log=self._log).tree



from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MismatchedTypes, NotSupportedOperation
from boa3.model.operation.binaryop import BinaryOp
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3_test.tests.boa_test import BoaTest


class TestArithmetic(BoaTest):

    def test_addition_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/Addition.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_subtraction_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.SUB
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/Subtraction.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_multiplication_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.MUL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/Multiplication.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_division_operation(self):
        path = '%s/boa3_test/test_sc/arithmetic_test/Division.py' % self.dirname
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_integer_division_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.DIV
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/IntegerDivision.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_modulo_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.MOD
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/Modulo.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_positive_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/Positive.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_negative_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.NEGATE
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/Negative.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_concatenation_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.CAT
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/Concatenation.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_power_operation(self):
        path = '%s/boa3_test/test_sc/arithmetic_test/Power.py' % self.dirname
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_str_multiplication_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + BinaryOp.StrMul.bytecode
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/StringMultiplication.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mismatched_type_binary_operation(self):
        path = '%s/boa3_test/test_sc/arithmetic_test/MismatchedOperandBinary.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_mismatched_type_unary_operation(self):
        path = '%s/boa3_test/test_sc/arithmetic_test/MismatchedOperandUnary.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_sequence_addition(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.PUSH4
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.LDARG0
            + Opcode.ADD
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/AdditionThreeElements.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mixed_operations(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x05'
            + Opcode.LDARG0
            + Opcode.LDARG2
            + Opcode.LDARG4
            + Opcode.MUL        # multiplicative operations
            + Opcode.ADD        # additive operations
            + Opcode.LDARG3
            + Opcode.NEGATE     # parentheses
            + Opcode.LDARG1
            + Opcode.DIV        # multiplicative
            + Opcode.SUB        # additive
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/MixedOperations.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mixed_operations_with_parentheses(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x05'
            + Opcode.LDARG0
            + Opcode.LDARG2
            + Opcode.LDARG4
            + Opcode.LDARG3
            + Opcode.NEGATE     # inside parentheses
            + Opcode.SUB        # parentheses
            + Opcode.MUL        # multiplicative operations
            + Opcode.LDARG1
            + Opcode.DIV        # multiplicative
            + Opcode.ADD        # additive operations
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/WithParentheses.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_addition_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.STARG0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/AdditionAugmentedAssignment.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_subtraction_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.SUB
            + Opcode.STARG0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/SubtractionAugmentedAssignment.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_multiplication_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.MUL
            + Opcode.STARG0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/MultiplicationAugmentedAssignment.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_division_augmented_assignment(self):
        path = '%s/boa3_test/test_sc/arithmetic_test/DivisionAugmentedAssignment.py' % self.dirname
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_integer_division_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.DIV
            + Opcode.STARG0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/IntegerDivisionAugmentedAssignment.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_modulo_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.MOD
            + Opcode.STARG0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/ModuloAugmentedAssignment.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_concatenation_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.CAT
            + Opcode.STARG0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/ConcatenationAugmentedAssignment.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_power_augmented_assignment(self):
        path = '%s/boa3_test/test_sc/arithmetic_test/PowerAugmentedAssignment.py' % self.dirname
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_str_multiplication_operation_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + BinaryOp.StrMul.bytecode
            + Opcode.STARG0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/StringMultiplicationAugmentedAssignment.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)



ONE_BYTE_MAX_VALUE = 255
TWO_BYTES_MAX_VALUE = 256 ** 2 - 1
FOUR_BYTES_MAX_VALUE = 256 ** 4 - 1

SIZE_OF_INT32 = 4
SIZE_OF_INT160 = 20
DEFAULT_UINT32 = 0

ENCODING = 'utf-8'
BYTEORDER = 'little'

INITIALIZE_METHOD_ID = '_initialize'



import ast
from abc import ABC, abstractmethod
from typing import Dict, List

from boa3.model.builtin.builtincallable import IBuiltinCallable
from boa3.model.expression import IExpression
from boa3.model.method import Method
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class IBuiltinDecorator(IBuiltinCallable, Method, ABC):
    def __init__(self, identifier: str, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None):
        super().__init__(identifier, args, defaults, return_type)

    @abstractmethod
    def validate_parameters(self, *params: IExpression) -> bool:
        """
        Verifies if the given parameters are valid to the method

        :param params: arguments of the method
        :return: True if all arguments are valid. False otherwise.
        """
        pass

    @property
    def requires_storage(self) -> bool:
        return False



from boa3.exception.CompilerError import (MetadataImplementationMissing, MetadataIncorrectImplementation,
                                          MismatchedTypes, MissingReturnStatement, UnexpectedArgument,
                                          UnresolvedReference)
from boa3.exception.CompilerWarning import RedeclaredSymbol
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3_test.tests.boa_test import BoaTest


class TestMetadata(BoaTest):

    def test_metadata_info_method(self):
        expected_output = (
            Opcode.PUSH5        # return 5
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoMethod.py' % self.dirname
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        # test features fields
        self.assertIn('features', manifest)
        self.assertIn('storage', manifest['features'])
        self.assertEqual(False, manifest['features']['storage'])
        self.assertIn('payable', manifest['features'])
        self.assertEqual(False, manifest['features']['payable'])

        # test extra field
        self.assertIn('extra', manifest)
        self.assertIsNone(manifest['extra'])

    def test_metadata_info_method_mismatched_type(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoMethodMismatchedReturn.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_metadata_info_method_no_return(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoMethodNoReturn.py' % self.dirname
        self.assertCompilerLogs(MissingReturnStatement, path)

    def test_metadata_info_multiple_method(self):
        expected_output = (
            Opcode.PUSH5        # return 5
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoMultipleMethod.py' % self.dirname
        self.assertCompilerLogs(RedeclaredSymbol, path)

        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('extra', manifest)
        self.assertIsNotNone(manifest['extra'])
        self.assertIn('Description', manifest['extra'])
        self.assertEqual('func1', manifest['extra']['Description'])

    def test_metadata_method_with_args(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataMethodWithArgs.py' % self.dirname
        self.assertCompilerLogs(UnexpectedArgument, path)

    def test_metadata_method_called_by_user_method(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataMethodCalledByUserMethod.py' % self.dirname
        self.assertCompilerLogs(UnresolvedReference, path)

    def test_metadata_object_call_user_method(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataObjectCallUserMethod.py' % self.dirname
        self.assertCompilerLogs(UnresolvedReference, path)

    def test_metadata_object_type_user_method(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataObjectTypeUserMethod.py' % self.dirname
        self.assertCompilerLogs(UnresolvedReference, path)

    def test_metadata_info_storage(self):
        expected_output = (
            Opcode.PUSH5        # return 5
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoStorage.py' % self.dirname
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('features', manifest)
        self.assertIn('storage', manifest['features'])
        self.assertEqual(True, manifest['features']['storage'])

    def test_metadata_info_storage_mismatched_type(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoStorageMismatchedType.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_metadata_info_payable(self):
        expected_output = (
            Opcode.PUSH5        # Main
            + Opcode.RET            # return 5
            + Opcode.PUSH0      # verify
            + Opcode.RET            # return False
        )

        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoPayable.py' % self.dirname
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('features', manifest)
        self.assertIn('storage', manifest['features'])
        self.assertEqual(True, manifest['features']['payable'])

        self.assertIn('abi', manifest)
        self.assertIn('methods', manifest['abi'])
        self.assertTrue(any(method['name'] == 'verify' for method in manifest['abi']['methods'] if 'name' in method))

    def test_metadata_info_payable_mismatched_type(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoPayableMismatchedType.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_metadata_info_payable_missing_verify(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoPayableMissingVerify.py' % self.dirname
        self.assertCompilerLogs(MetadataImplementationMissing, path)

    def test_metadata_info_payable_incorrect_verify(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoPayableIncorrectVerify.py' % self.dirname
        self.assertCompilerLogs(MetadataIncorrectImplementation, path)

    def test_metadata_info_payable_private_verify(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoPayablePrivateVerify.py' % self.dirname
        self.assertCompilerLogs(MetadataIncorrectImplementation, path)

    def test_metadata_info_author(self):
        expected_output = (
            Opcode.PUSH5        # return 5
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoAuthor.py' % self.dirname
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('extra', manifest)
        self.assertIsNotNone(manifest['extra'])
        self.assertIn('Author', manifest['extra'])
        self.assertEqual('Test', manifest['extra']['Author'])

    def test_metadata_info_author_mismatched_type(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoAuthorMismatchedType.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_metadata_info_email(self):
        expected_output = (
            Opcode.PUSH5        # return 5
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoEmail.py' % self.dirname
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('extra', manifest)
        self.assertIsNotNone(manifest['extra'])
        self.assertIn('Email', manifest['extra'])
        self.assertEqual('test@test.com', manifest['extra']['Email'])

    def test_metadata_info_email_mismatched_type(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoEmailMismatchedType.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_metadata_info_description(self):
        expected_output = (
            Opcode.PUSH5        # return 5
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoDescription.py' % self.dirname
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('extra', manifest)
        self.assertIsNotNone(manifest['extra'])
        self.assertIn('Description', manifest['extra'])
        self.assertEqual('This is an example', manifest['extra']['Description'])

    def test_metadata_info_description_mismatched_type(self):
        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoDescriptionMismatchedType.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_metadata_info_extras(self):
        expected_output = (
            Opcode.PUSH5        # return 5
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/metadata_test/MetadataInfoExtras.py' % self.dirname
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('extra', manifest)
        self.assertIsNotNone(manifest['extra'])
        self.assertIn('Author', manifest['extra'])
        self.assertEqual('Test', manifest['extra']['Author'])
        self.assertIn('Email', manifest['extra'])
        self.assertEqual('test@test.com', manifest['extra']['Email'])
        self.assertIn('Description', manifest['extra'])
        self.assertEqual('This is an example', manifest['extra']['Description'])



from __future__ import annotations

import base64
import binascii
import json
from enum import IntFlag
from typing import List, Callable

from boa3.neo3 import contracts
from boa3.neo3.core import serialization, types, IJson, cryptography, utils
from boa3.neo3.core.serialization import BinaryReader, BinaryWriter


class ContractGroup(IJson):
    """
    Describes a set of mutually trusted contracts.

    See Also: ContractManifest.
    """

    def __init__(self, public_key: cryptography.EllipticCurve.ECPoint, signature: bytes):
        self.public_key = public_key
        self.signature = signature

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.public_key == other.public_key and self.signature == other.signature

    def is_valid(self, contract_hash: types.UInt160) -> bool:
        """
        Validate if the group has agreed on allowing the specific contract_hash.

        Args:
            contract_hash:
        """
        return cryptography.verify_signature(contract_hash.to_array(),
                                             self.signature,
                                             self.public_key.encode_point(False))

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = {
            'pubKey': str(self.public_key),
            'signature': base64.b64encode(self.signature).decode()
        }
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractGroup:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
        """
        return cls(
            public_key=cryptography.EllipticCurve.ECPoint.deserialize_from_bytes(binascii.unhexlify(json['pubKey'])),
            signature=base64.b64decode(json['signature'].encode('utf8'))
        )


class ContractPermission(IJson):
    """
    An object describing a single set of outgoing call restrictions for a 'System.Contract.Call' SYSCALL.
    It describes what other smart contracts the executing contract is allowed to call and what exact methods on the
    other contract are allowed to be called. This is enforced during runtime.

    Example:
        Contract A (the executing contract) wants to call method "x" on Contract B. The runtime will query the manifest
        of Contract A and ask if this is allowed. The Manifest will search through its permissions (a list of
        ContractPermission objects) and ask if this "is_allowed(target_contract, target_method)".
    """

    def __init__(self, contract: contracts.ContractPermissionDescriptor, methods: WildcardContainer):
        self.contract = contract
        self.methods = methods

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.contract == other.contract and self.methods == other.methods

    @classmethod
    def default_permissions(cls) -> ContractPermission:
        """
        Construct a ContractPermission which allows any contract and any method to be called.
        """
        return cls(contracts.ContractPermissionDescriptor(),  # with no parameters equals to Wildcard
                   WildcardContainer.create_wildcard())

    def is_allowed(self, manifest: ContractManifest, method: str) -> bool:
        """
        Return if it is allowed to call `method` on contract `manifest.contract_hash`.

        Args:
            manifest: the manifest of the contract to be called.
            method: the method of the contract to be called.
        """
        if self.contract.is_hash:
            if not self.contract.contract_hash == manifest.contract_hash:
                return False
        elif self.contract.is_group:
            results = list(map(lambda p: p.public_key != self.contract.group, manifest.groups))
            if all(results):
                return False
        return self.methods.is_wildcard or method in self.methods

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = self.contract.to_json()
        # because NEO C# returns a string from "method" instead of sticking to a standard interface
        json.update({'methods': self.methods.to_json()['wildcard']})
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractPermission:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
        """
        cpd = contracts.ContractPermissionDescriptor.from_json(json)
        json_wildcard = {'wildcard': json['methods']}
        methods = WildcardContainer.from_json(json_wildcard)
        return cls(cpd, methods)


class ContractFeatures(IntFlag):
    NO_PROPERTY = 0,
    #: Indicate the contract has storage.
    HAS_STORAGE = 1 << 0
    #: Indicate the contract accepts tranfers.
    PAYABLE = 1 << 2


class WildcardContainer(IJson):
    """
    An internal helper class for ContractManifest attributes.
    """

    def __init__(self, data: list = None):
        self._is_wildcard = False
        self._data = data if data else []

    def __contains__(self, item):
        return item in self._data

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self):
        return len(self._data)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self._data == other._data

    @classmethod
    def create_wildcard(cls) -> WildcardContainer:
        """
        Creates an instance that indicates any value is allowed.
        """
        instance = cls()
        instance._is_wildcard = True
        return instance

    @property
    def is_wildcard(self) -> bool:
        """
        Indicates if the container is configured to allow all values.
        """
        return self._is_wildcard

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        if self.is_wildcard:
            return {'wildcard': '*'}
        return {'wildcard': list(map(lambda d: str(d), self._data))}

    @classmethod
    def from_json(cls, json: dict):
        """
        Parse object out of JSON data.

        Note: if the value is not '*', and is a Python list, then it will assume
        that the list members are strings or convertible via str().

        If the wildcard should contain other data types, use the alternative `from_json_as_type()` method

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
            ValueError: if the data supplied cannot recreate a valid object.
        """
        value = json.get('wildcard', None)
        if value is None:
            raise ValueError(f"Invalid JSON - Cannot recreate wildcard from None")
        if value == '*':
            return WildcardContainer.create_wildcard()
        if isinstance(value, list):
            return WildcardContainer(data=list(map(lambda d: str(d), value)))
        raise ValueError(f"Invalid JSON - Cannot deduce WildcardContainer type from: {value}")

    @classmethod
    def from_json_as_type(cls, json: dict, conversion_func: Callable):
        """
            Parse object out of JSON data.

            Note: if the value is not '*', and is a Python list, then it will use `conversion_func` to
            parse the members into the expected types.

        Args:
            json: a dictionary.
            conversion_func: a callable that takes 1 argument, which is the element in the value list

            Example with UInt160:
                {'wildcard': ['0xa400ff00ff00ff00ff00ff00ff00ff00ff00ff01', 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa']}

                 the first call has as argument '0xa400ff00ff00ff00ff00ff00ff00ff00ff00ff01'.
                 to process this example call
                 WildcardContainer.from_json_as_type(json_data, lambda f: types.UInt160.from_string(f))

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
            ValueError: if the data supplied cannot recreate a valid object.
        """
        value = json.get('wildcard', None)
        if value is None:
            raise ValueError(f"Invalid JSON - Cannot recreate wildcard from None")
        if value == '*':
            return WildcardContainer.create_wildcard()
        if isinstance(value, list):
            return WildcardContainer(data=list(map(lambda d: conversion_func(d), value)))
        raise ValueError(f"Invalid JSON - Cannot deduce WildcardContainer type from: {value}")


class ContractManifest(serialization.ISerializable, IJson):
    """
    A description of a smart contract's abilities (callable methods & events) as well as a set of restrictions
    describing what external contracts and methods are allowed to be called.

    For more information see:
    https://github.com/neo-project/proposals/blob/3e492ad05d9de97abb6524fb9a73714e2cdc5461/nep-15.mediawiki
    """
    #: The maximum byte size after serialization to be considered valid a valid contract.
    MAX_LENGTH = 2048

    def __init__(self, contract_hash: types.UInt160 = types.UInt160.zero()):
        """
        Creates a default contract manifest if not arguments are supplied.

        A default contract is not Payable and has no storage as configured by its features.
        It may not be called by any other contracts

        Args:
            contract_hash: the contract script hash to create a manifest for.
        """
        #: A group represents a set of mutually trusted contracts. A contract will trust and allow any contract in the
        #: same group to invoke it.
        self.groups: List[ContractGroup] = []

        #: Features describe what contract abilities are available. TODO: link to contract features
        self.features: ContractFeatures = ContractFeatures.NO_PROPERTY

        #: For technical details of ABI, please refer to NEP-14: NeoContract ABI.
        #: https://github.com/neo-project/proposals/blob/d1f4e9e1a67d22a5755c45595121f80b0971ea64/nep-14.mediawiki
        self.abi: contracts.ContractABI = contracts.ContractABI(
            contract_hash=contract_hash,
            entry_point=contracts.ContractMethodDescriptor.default_entrypoint(),
            events=[],
            methods=[]
        )

        #: Permissions describe what external contract(s) and what method(s) on these are allowed to be invoked.
        self.permissions: List[contracts.ContractPermission] = [contracts.ContractPermission.default_permissions()]

        self.contract_hash: types.UInt160 = self.abi.contract_hash

        # Update trusts/safe_methods with outcome of https://github.com/neo-project/neo/issues/1664
        # Unfortunately we have to add this nonsense logic or we get deviating VM results.
        self.trusts = WildcardContainer()  # for UInt160 types
        self.safe_methods = WildcardContainer()  # for string types
        self.extra = None

    def __len__(self):
        return utils.get_var_size(str(self.to_json()).replace(' ', ''))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.groups == other.groups
                and self.features == other.features
                and self.abi == other.abi
                and self.permissions == other.permissions
                and self.trusts == other.trusts
                and self.safe_methods == other.safe_methods
                and self.extra == other.extra)

    def serialize(self, writer: BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_var_string(json.dumps(self.to_json()).replace(' ', ''))

    def deserialize(self, reader: BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self._deserialize_from_json(json.loads(reader.read_var_string(self.MAX_LENGTH)))

    def _deserialize_from_json(self, json: dict) -> None:
        self.abi = contracts.ContractABI.from_json(json['abi'])
        self.groups = list(map(lambda g: ContractGroup.from_json(g), json['groups']))
        self.features = ContractFeatures.NO_PROPERTY
        if json['features']['storage']:
            self.features |= ContractFeatures.HAS_STORAGE
        if json['features']['payable']:
            self.features |= ContractFeatures.PAYABLE
        self.permissions = list(map(lambda p: ContractPermission.from_json(p), json['permissions']))

        self.trusts = WildcardContainer.from_json_as_type(
            {'wildcard': json['trusts']},
            lambda t: types.UInt160.from_string(t))

        # converting json key/value back to default WildcardContainer format
        self.safe_methods = WildcardContainer.from_json({'wildcard': json['safeMethods']})
        self.extra = json['extra']

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        trusts = list(map(lambda m: "0x" + m, self.trusts.to_json()['wildcard']))
        json = {
            "groups": list(map(lambda g: g.to_json(), self.groups)),
            "features": {
                "storage": contracts.ContractFeatures.HAS_STORAGE in self.features,
                "payable": contracts.ContractFeatures.PAYABLE in self.features,
            },
            "abi": self.abi.to_json(),
            "permissions": list(map(lambda p: p.to_json(), self.permissions)),
            "trusts": trusts,
            "safeMethods": self.safe_methods.to_json()['wildcard'],
            "extra": self.extra
        }
        return json

    @classmethod
    def from_json(cls, json: dict):
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raise:
            KeyError: if the data supplied does not contain the necessary keys.
        """
        instance = cls(types.UInt160.zero())
        instance._deserialize_from_json(json)
        return instance

    def is_valid(self, contract_hash: types.UInt160) -> bool:
        """
        Test the manifest data for correctness and return the result.

        - Compares the manifest ABI with the `contract_hash`
        - Validates the manifest groups with the `contract_hash`

        An example use-case is to create and update smart contracts in a safe manner on the chain.

        Args:
            contract_hash:

        """
        if not self.abi.contract_hash == contract_hash:
            return False
        result = list(map(lambda g: g.is_valid(contract_hash), self.groups))
        return all(result)

    def can_call(self, target_manifest: ContractManifest, method: str) -> bool:
        """
        Convenience function to check if it is allowed to call `method` on `target_manifest` from this contract.

        Args:
            target_manifest: The manifest describing the target contract.
            method: the name of the target method.
        """
        results = list(map(lambda p: p.is_allowed(target_manifest, method), self.permissions))
        return any(results)



from typing import Any

from boa3.builtin.interop.runtime.triggertype import TriggerType


def check_witness(hash_or_pubkey: bytes) -> bool:
    """
    Verifies that the transactions or block of the calling contract has validated the required script hash.

    :param hash_or_pubkey: script hash or public key to validate
    :type hash_or_pubkey: bytes
    :return: a boolean value that represents whether the script hash was verified.
    :rtype: bool
    """
    pass


def notify(state: Any):
    """
    Notifies the client from the executing smart contract.

    :param state: the notification message
    """
    pass


def log(message: str):
    """
    Show log messages to the client from the executing smart contract.

    :param message: the log message
    :type message: str
    """
    pass


def trigger() -> TriggerType:
    """
    Verifies if the smart contract trigger is an application trigger.

    :return: a boolean value that represents whether the contract trigger is an application.
    :rtype: TriggerType
    """
    pass


def is_application_trigger() -> bool:
    """
    Verifies if the smart contract trigger is an application trigger.

    :return: a boolean value that represents whether the contract trigger is an application.
    :rtype: bool
    """
    pass


def is_verification_trigger() -> bool:
    """
    Verifies if the smart contract trigger is an verification trigger.

    :return: a boolean value that represents whether the contract trigger is a verification.
    :rtype: bool
    """
    pass


calling_script_hash: bytes = b''



from typing import Any, Set

from boa3.model.type.collection.mapping.mappingtype import MappingType
from boa3.model.type.itype import IType


class GenericMappingType(MappingType):
    """
    An class used to represent a generic Python sequence type
    """

    def __init__(self, keys_type: Set[IType] = None, values_type: Set[IType] = None):
        identifier: str = 'mapping'
        keys_type = self.filter_types(keys_type)
        values_type = self.filter_types(values_type)
        super().__init__(identifier, keys_type, values_type)

    def is_valid_key(self, key_type: IType) -> bool:
        return key_type == self.valid_key

    @property
    def is_generic(self) -> bool:
        return True

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, MappingType)

    def __hash__(self):
        return hash(self.identifier)



from boa3_test.tests.boa_test import BoaTest


class TestTemplate(BoaTest):

    def test_nep5(self):
        path = '%s/boa3_test/examples/nep5.py' % self.dirname
        output, manifest = self.compile_and_save(path)



from typing import Any, Set

from boa3.model.type.collection.mapping.mutable.mutablemappingtype import MutableMappingType
from boa3.model.type.itype import IType


class DictType(MutableMappingType):
    """
    A class used to represent Python dict type
    """

    def __init__(self, keys_type: Set[IType] = None, values_type: Set[IType] = None):
        identifier = 'dict'
        keys_type = self.filter_types(keys_type)
        values_type = self.filter_types(values_type)
        super().__init__(identifier, keys_type, values_type)

    @property
    def default_value(self) -> Any:
        return dict()

    @classmethod
    def _is_type_of(cls, value: Any):
        return type(value) in [dict, DictType]

    def __hash__(self):
        return hash(self.identifier)
